#!/usr/bin/env python3

"""
** Manages all client/server tcp communication. **
--------------------------------------------------

This is where the main classes representing the client and the servers are defined.
The verifications and the communication protocols are also defined here.
"""

__all__ = ['Client', 'Server']


from raisin.communication.client import Client
from raisin.communication.server import Server



#!/usr/bin/env python3

"""
** Allows for extensive testing around communication. **
--------------------------------------------------------
"""

import socket

from raisin.communication.abstraction import SocketAbstractConn

def test_socket_abstraction_conn():
    """
    ** Tests all fonctions of the abstraction. **
    """
    socket1, socket2 = socket.socketpair()
    abstraction1, abstraction2 = SocketAbstractConn(socket1), SocketAbstractConn(socket2)

    abstraction1.send((b'mes', b'sage1',))
    abstraction2.send((b'hello',))
    abstraction2.send((b'',))
    abstraction1.send((b'message2',))
    assert b''.join(abstraction2.recv()) == b'message1'
    assert b''.join(abstraction2.recv()) == b'message2'
    assert b''.join(abstraction1.recv()) == b'hello'
    assert b''.join(abstraction1.recv()) == b''

    abstraction1.send_obj(0)
    abstraction1.send_obj(1)
    abstraction2.send_obj(2)
    abstraction2.send_obj(3)
    assert abstraction1.recv_obj() == 2
    assert abstraction1.recv_obj() == 3
    assert abstraction2.recv_obj() == 0
    assert abstraction2.recv_obj() == 1

    abstraction1.close()
    abstraction2.close()



#!/usr/bin/env python3

"""
** Makes the communication medium and its use independent. **
-------------------------------------------------------------

Whether it is tcp sockets, files, mailboxes or a drive like dropbox,
we want to be able to use all these media indifferently...
That's the role of this module.
It allows to add an abstraction layer to unify the communication API.
"""

import abc
import socket

from raisin.serialization import deserialize, serialize
from raisin.serialization.iter_tools import anticipate
from raisin.serialization.core import size2tag, tag2size


class AbstractConn:
    """
    ** Allows for easier communication. **

    It is the basic abstract class that allows to make information
    carriers of very different nature communicate in the same way.
    """

    @abc.abstractmethod
    def send(self, raw, dest=None):
        """
        ** Sends a data stream to the recipient. **

        Must be overwritten.

        Parameters
        ----------
        raw : iterator
            A generator that yields byte strings. The division is arbitrary.
            On reception, the packets received are not necessarily split in the same way.
            This flow can typically be generated by the ``raisin.serialization.serialize`` function.
        dest : str, default=None
            The address or addresses of the recipient(s) of the message.
            By default (None), the message is sent to the main recipient.
        """
        raise NotImplementedError('this abstract method must be overwritten')

    def send_obj(self, obj, dest=None, **kwargs):
        """
        ** Sends an object to the recipient. **

        Can be overwritten.

        Parameters
        ----------
        obj
            It is the serialisable object that is sent to the recipient(s).
        dest : str, default=None
            Same as ``AbstractConn.send``.
        **kwargs : dict
            These are the parameters passed to the ``raisin.serialization.serialize`` function.
        """
        self.send(serialize(obj, **kwargs), dest=dest)

    @abc.abstractmethod
    def recv(self):
        """
        ** Recovers the raw data sent by the sender. **

        Must be overwritten.

        Yields
        ------
        bytes
            The packets of bytes emitted by the ``AbstractConn.send``
            method of the sender. Packet decoding may differ, but the order is preserved.

        raises
        ------
        ConnectionError
            If the connection is broken.
        """
        raise NotImplementedError('this abstract method must be overwritten')

    def recv_obj(self, **kwargs):
        """
        ** Recover and deserialize the sent object. **

        Can be overwritten.

        Parameters
        ----------
        **kwargs : dict
            These are the parameters passed to the ``raisin.serialization.deserialize`` function.

        Returns
        -------
        obj
            The emitted object.
        """
        return deserialize(self.recv(), **kwargs)

    @abc.abstractmethod
    def close(self):
        """
        ** Definitely close the connection. **

        Must be overwritten.
        """
        raise NotImplementedError('this abstract method must be overwritten')

    def __del__(self):
        self.close()


class SocketAbstractConn(AbstractConn):
    """
    ** Adds an abstraction layer to TCP sockets. **

    Attributes
    ----------
    conn : socket.socket
        A socket capable of reading and sending data.
    """

    def __init__(self, conn):
        self.conn = conn
        self._recv_pack = b'' # buffer memory

    def send(self, raw, dest=None):
        """
        ** Implementation of ``AbstactConn.send``. **
        """
        for is_end, pack in anticipate(raw):
            signed_pack = bytes([is_end]) + size2tag(len(pack)) + pack
            self.conn.sendall(signed_pack)

    def recv(self):
        """
        ** Implementation of ``AbstraConn.recv``. **
        """
        def gen():
            while True:
                try:
                    yield self.conn.recv(4096)
                except OSError as err:
                    raise ConnectionError('communication seems to be interrupted') from err
        iter_pack = iter(gen())
        while True:
            # check for len and last pack
            while not self._recv_pack:
                self._recv_pack += next(iter_pack)
            is_end = bool(self._recv_pack[0])
            self._recv_pack = self._recv_pack[1:]
            size, self._recv_pack, _ = tag2size(pack=self._recv_pack, gen=iter_pack)
            while len(self._recv_pack) < size:
                self._recv_pack += next(iter_pack)
            yield self._recv_pack[:size]
            self._recv_pack = self._recv_pack[size:]
            if is_end:
                break

    def close(self):
        """
        ** Implementation of ``AbstractConn.close``. **
        """
        try:
            self.conn.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        self.conn.close()



#!/usr/bin/env python3

"""
** Manages the serialization of python objects. **
--------------------------------------------------

This is where we take care of bijectively converting abstract python objects
into a sequence of bytes.
This is also where we take care of data compression and encryption.

Notes
-----
The functions accessible here are high-level APIs that do the input checks.
For a more intensive use where the inputs are checked,
it is better to go directly to the 'low level' functions.
"""

import re

__all__ = ['deserialize', 'dump', 'dumps', 'load', 'loads', 'serialize']


def deserialize(data, **kwargs):
    """
    ** Deserializes the encoded object by a *raisin* serialization function. **

    Parameters
    ----------
    data
        Data from a serialization function. They can take different forms:

        - *bytes* : Recommended for small objects.
            For large objects, RAM may be a limiting factor.
        - *generator of bytes* : This is the recommended form because it is the most effective.
            It allows to minimize the calculations while preserving the RAM.
            For example it can be the output of the ``serialize`` function.
        - *str* : Not recommended because it is not very effective.
            It's better if you do the conversion to bytes yourself.
        - *file like* : Must have a read() method that takes
            an integer argument and returns bytes or str. Thus *file* can be a
            binary file object opened for reading, an io.BytesIO object, or any
            other custom object that meets this interface.
    psw: str or bytes or None, default=None
        - *str* : Password for symmetric AES deciphering.
        - *bytes* : Public RSA key in PEM format.
        - *None* : Do not attempt any deciphering.
    paralleling_rate : int, default=0
        Allows you to parrallelize the calculations :

        - *0* : No parallelization, everything is executed in the current process.
        - *1* : Slight parallelization that creates several threads with the *threading* modules.
        - *2* : Parallelization that uses the different cores of the machine
            with the 'multiprocessing' module.

    Returns
    -------
    object
        The initially serialized python object.
        It is not the same instance as the initial object but the returned object
        is equivalent to the initial one.

    Examples
    --------
    >>> from raisin.serialization import deserialize
    >>> deserialize(b'/small int/123456789')
    123456789
    >>>
    """
    kwargs['psw'] = kwargs.get('psw', None)
    kwargs['paralleling_rate'] = kwargs.get('paralleling_rate', 0)

    assert isinstance(data, (bytes, str)) or hasattr(data, '__iter__') or hasattr(data, 'read')
    assert isinstance(kwargs['paralleling_rate'], int)
    assert 0 <= kwargs['paralleling_rate'] <= 2

    if kwargs['psw'] is not None:
        if isinstance(kwargs['psw'], bytes):
            try:
                if not re.search(
                    r'-----BEGIN RSA PRIVATE KEY-----(.|\n)+-----END RSA PRIVATE KEY-----',
                    kwargs['psw'].decode(),
                ):
                    raise ValueError('The key is not a private key in PEM format.')
            except UnicodeDecodeError as err:
                raise ValueError(
                    'If it is of type (bytes), the password must be an RSA '
                    'private key in PEM format, this is not the case.'
                ) from err
        elif isinstance(kwargs['psw'], str):
            if kwargs['psw'] == '':
                raise ValueError(
                    'The password should not be an empty string, '
                    'if there is no password, it should not be specified.'
                )
        else:
            raise TypeError("'psw' have to be str or bytes.")

    from raisin.serialization.iter_tools import data2gen

    pack, gen = data2gen(data)
    from raisin.serialization.core import deserialize as deserialize_

    return deserialize_(pack, gen, **kwargs)


def dump(obj, file, **kwargs):
    """
    ** Write a representation of obj to the open file object file. **

    This function is an overlay of the ``serialize`` function,
    but may be more efficient.

    Parameters
    ----------
    obj
        The object to serialize.
    file
        The *file* argument must have a write() method that accepts a single
        bytes argument. It can thus be a file object opened for binary
        writing, an io.BytesIO instance, or any other custom object that meets
        this interface.
    **kwargs : dict, optional
        These are the arguments directly transferred to the ``serialize`` function.
        Only the 'compresslevel' parameter differs, because here the default value is 5.

    See Also
    --------
    serialize : Main function for 'low level' serialization.

    Examples
    --------
    >>> import tempfile
    >>> from raisin.serialization import dump
    >>> with tempfile.TemporaryFile('wb') as file:
    ...     dump(123456789, file)
    ...
    >>>
    """
    assert hasattr(file, 'write')
    for pack in serialize(obj, **kwargs):
        file.write(pack)


def dumps(obj, **kwargs):
    """
    ** Serialize an object and return a nice string. **

    This function is an overlay of the ``serialize`` function.
    Instead of yielding a string of bytes, this function
    retrieves a simple printable ascii string.
    The number of characters given by this function tends
    asymptotically to be 64/51 larger than the number of bytes
    given by the ``serialize`` function.

    Parameters
    ----------
    obj
        The object to serialize.
    **kwargs : dict, optional
        These are the arguments directly transferred to the ``serialize`` function.
        Only the 'compresslevel' parameter differs, because here the default value is 1.

    Returns
    -------
    str
        A printable ascii character string. This string can be used as input
        parameter to the 'loads' function. But this protocol is also compatible
        with the 'load' and 'deserialize' functions.

    See Also
    --------
    serialize : Main function for 'low level' serialization.
    dump : Allows to serialize efficiently in a file.

    Examples
    --------
    >>> from raisin.serialization import dumps
    >>> dumps(123456789)
    '00aWDBam0vc'
    >>>
    """
    kwargs['compresslevel'] = kwargs.get('compresslevel', 1)
    from raisin.serialization.core import bytes2str
    from raisin.serialization.constants import HEADER

    return HEADER['dumps'][1].decode(encoding='ascii') + bytes2str(serialize(obj, **kwargs))


def load(file, **kwargs):
    """
    ** Bijection of the 'dump' function. **

    Allows to reconstitute aserialized object in a file.

    Parameters
    ----------
    file
        The argument *file* must have a read() method that takes
        an integer argument and returns bytes or str. Thus *file* can be a
        binary file object opened for reading, an io.BytesIO object, or any
        other custom object that meets this interface.
    **kwargs : dict
        These are the arguments directly transferred to the ``deserialize`` function.

    Returns
    -------
    object
        The initially serialized python object.
        It is not the same instance as the initial object but the returned object
        is equivalent to the initial one.

    Examples
    --------
    >>> from os.path import join
    >>> from tempfile import mkdtemp
    >>> from raisin.serialization import dump, load
    >>>
    >>> filepath = join(mkdtemp(), 'object.rsn')
    >>> with open(filepath, 'wb') as file:
    ...     dump(123456789, file)
    ...
    >>> with open(filepath, 'rb') as file:
    ...     obj = load(file)
    ...
    >>> obj
    123456789
    >>>
    """
    assert hasattr(file, 'read')
    return deserialize(file, **kwargs)


def loads(string, **kwargs):
    """
    ** Bijection of the 'dumps' function. **

    Allows to reconstitute an object serialized by the 'dumps' function.

    Parameters
    ----------
    string : str
        The ascii string from the 'dumps' function.
    **kwargs : dict
        These are the arguments directly transferred to the ``deserialize`` function.

    Returns
    -------
    object
        The initially serialized python object.
        It is not the same instance as the initial object but the returned object
        is equivalent to the initial one.

    Examples
    --------
    >>> from raisin.serialization import loads
    >>> loads('00aWDBam0vc')
    123456789
    >>>
    """
    assert isinstance(string, str)
    return deserialize(string.encode('ascii'), **kwargs)


def serialize(obj, **kwargs):
    """
    ** Main function to serialize a python object. **

    Unlike 'pickle', this function supports a wider variety of objects.
    It serializes in real time by working on batches of data.
    RAM is therefore highly preserved even when multi-GB objects are serialized.

    Parameters
    ----------
    obj
        The object to serialize.
    compresslevel : int, default=0
        Allows lossless compression of the result in order to reduce the size
        of the transferred data. Can take the following values :

        - *-1* : Adaptive compression rate that provides optimal compression
            in a rate-limited communication channel.
            Works only if *parralelization_rate* > 0.
        - *0* : There is no compression.
        - *1* : Chose more compact tags but more difficult to interpret by a human.
            This is not a compression that tries to eliminate redundancy but only
            tries not to add more.
        - *2* : Light but fast compression.
            It is a good compromise between the time lost and the data rate gained.
        - *3* : Strong enough compression that aims for a compression ratio
            equal to 95% of the maximum compression ratio.
        - *4* : Maximum compression that does not care about the amount of resources.
    copy_file : boolean, default=True
        In case *obj* is a path_like to a file, the file itself is serialized (if True).
        Whatever the value of *copy_file*, *obj* remains unchanged.
    psw : str or bytes or None, default=None
        Allows you to ciffer the return value :

        - *str* : Symmetric ciphering with the AES algorithm.
        - *bytes* : Public key in PEM format, RSA is used.
        - *None* : No ciphering.
    authenticity : boolean, default=False
        Adds a chriptographic hash (if True). This ensures that the data is authentic.
    paralleling_rate : int, default=0
        Allows you to parrallelize the calculations :

        - *0* : No parallelization, everything is executed in the current process.
        - *1* : Slight parallelization that creates several threads with the *threading* modules.
        - *2* : Parallelization that uses the different cores of the machine
            with the 'multiprocessing' module.

    Yields
    ------
    bytes
        The concatenation of all ceded packages contains sufficient
        information to reconstruct the original object.

    Examples
    --------
    >>> from raisin.serialization import serialize
    >>>
    """
    kwargs['compresslevel'] = kwargs.get('compresslevel', 0)
    kwargs['copy_file'] = kwargs.get('copy_file', True)
    kwargs['psw'] = kwargs.get('psw', None)
    kwargs['authenticity'] = kwargs.get('authenticity', (kwargs['psw'] is not None))
    kwargs['paralleling_rate'] = kwargs.get('paralleling_rate', 0)

    assert isinstance(kwargs['compresslevel'], int)
    assert isinstance(kwargs['copy_file'], bool)
    assert isinstance(kwargs['authenticity'], bool)
    assert isinstance(kwargs['paralleling_rate'], int)
    if kwargs['psw'] is not None:
        if isinstance(kwargs['psw'], bytes):
            try:
                if not re.search(
                    r'-----BEGIN PUBLIC KEY-----(.|\n)+-----END PUBLIC KEY-----',
                    kwargs['psw'].decode(),
                ):
                    raise ValueError('The key is not a public key in PEM format.')
            except UnicodeDecodeError as err:
                raise ValueError(
                    'If it is of type (bytes), the password must be an RSA public '
                    'key in PEM format, this is not the case'
                ) from err
        elif isinstance(kwargs['psw'], str):
            if kwargs['psw'] == '':
                raise ValueError(
                    'The password should not be an empty string, '
                    'if there is no password, it should not be specified.'
                )
        else:
            raise TypeError("'psw' have to be str or bytes.")
    assert -1 <= kwargs['compresslevel'] <= 4
    assert 0 <= kwargs['paralleling_rate'] <= 2

    from raisin.serialization.core import serialize

    yield from serialize(obj, **kwargs)



#!/usr/bin/env python3

"""
** Offers small and simple tools to help manage generators. **
--------------------------------------------------------------

More specifically, these are functions dedicated to flow management
for serialization and deserialization.
"""

import itertools
import tempfile

from raisin.serialization.constants import BUFFER_SIZE


class _Null:
    """
    Stratagem to find out if a default setting is provided.
    """


def anticipate(gen):
    """
    ** Allows to know if the ceded packet is the last one. **

    Parametres
    ----------
    gen : iterable
        Generator of unknown 'length', which gives away objects of unknown length.

    Yields
    ------
    is_end : boolean
        A boolean which is True if the packet is the last one, False otherwise.
    pack
        The packet given up by the input generator.

    Raises
    ------
    RuntimeError
        In case there are no packages to count.

    Notes
    -----
    There is no verification on the entries.

    Examples
    --------
    >>> from raisin.serialization.iter_tools import anticipate
    >>>
    >>> for is_end, pack in anticipate(range(3)):
    ...     print(is_end, pack)
    ...
    False 0
    False 1
    True 2
    >>> list(anticipate(['pack']))
    [(True, 'pack')]
    >>>
    """
    is_empty = True
    current = None

    for i, current in enumerate(gen):
        if i == 0:
            previous = current
            is_empty = False
            continue
        yield False, previous
        previous = current

    if is_empty:
        raise RuntimeError('The generator must not be empty.')
    yield True, current


def concat_gen(gen):
    r"""
    ** Add flags so that you can find the current division. **

    Goes together with the ``deconcat_gen`` function.

    Parameters
    ----------
    gen : generator
        Bytes string generator.

    Yields
    ------
    bytes
        The same packet as the input generator with a flag that
        gives the length of the packet.

    Examples
    --------
    >>> from raisin.serialization.iter_tools import concat_gen
    >>>
    >>> gen = [b'a', b'houla', b'', b'hihi']
    >>> b''.join(concat_gen(gen))
    b'\x81a\x85houla\x80\x84hihi'
    >>>
    """
    from raisin.serialization.core import size2tag

    yield from (size2tag(len(e)) + e for e in gen)


def data2gen(data):
    """
    ** Transforms the input into a byte generator. **

    Parameters
    ----------
    data
        A serialized object that can take various forms:

        - *bytes* : The entrance is returned as it was.
        - *generator* : The generator must yield strings of bytes or characters.
        - *BufferedReader* : The content of the file is transferred in packages.
        - *str* : (not recommended) The string must be encoded in ascii.
            The binary representation of the string is returned.
        - *TextIOWrapper* : (not recommended) The ascii content of the
            file is transferred in packages.

    Returns
    -------
    bytes
        A packet of bytes that precedes the generator.
    generator
        A byte chain generator.

    Raises
    ------
    ValueError
        In case a string is not written entirely in ascii.
    TypeError
        If the input is not convertible into a byte generator.

    Examples
    --------
    >>> from raisin.serialization.iter_tools import data2gen
    >>>
    >>> pack, gen = data2gen(b'bytes string')
    >>> pack, list(gen)
    (b'bytes string', [])
    >>> pack, gen = data2gen('str string')
    >>> pack, list(gen)
    (b'str string', [])
    >>> pack, gen = data2gen(iter([b'a generator']))
    >>> pack, list(gen)
    (b'', [b'a generator'])
    >>>
    """

    def _file_like(data):
        while True:
            pack = data.read(BUFFER_SIZE)
            if pack:
                if isinstance(pack, bytes):
                    yield pack
                else:
                    try:
                        yield pack.encode(encoding='ascii')
                    except UnicodeEncodeError as err:
                        raise ValueError(
                            'If you want to deserialize a txt file '
                            '(which is not recommended), this file '
                            'must be written entirely in ascii.'
                        ) from err
            else:
                break

    def _iter(data):
        for pack in data:
            if isinstance(pack, bytes):
                yield pack
            elif isinstance(pack, str):
                try:
                    yield pack.encode(encoding='ascii')
                except UnicodeEncodeError as err:
                    raise ValueError(
                        'If you want to deserialize a string generator. '
                        'The strings must be written entirely in ascii.'
                    ) from err
            else:
                raise TypeError(
                    'The packets given by the generator must be of type '
                    f'str or bytes, not {pack.__class__.__name__}.'
                )

    if isinstance(data, bytes):
        return data, (lambda: (yield from []))()
    if isinstance(data, str):
        try:
            return data.encode(encoding='ascii'), (lambda: (yield from []))()
        except UnicodeEncodeError as err:
            raise ValueError(
                'if you want to deserialize a string, the string '
                'must be written entirely in ascii.'
            ) from err
    elif hasattr(data, 'read'):
        return b'', _file_like(data)
    elif hasattr(data, '__iter__'):
        return b'', _iter(data)
    else:
        raise TypeError('The argument provided is not in a supported type.')


def deconcat_gen(*, pack=b'', gen=(lambda: (yield from []))()):
    r"""
    ** Reconcile and re-cut the packets to their original shape. **

    Goes together with the ``concat_gen`` function.

    Parameters
    ----------
    pack : bytes
        The first element of the generator.
    gen : generator
        The rest of the packages to be rearranged.

    Yields
    ------
    bytes
        The reordered packages like the ones that were once provided
        at the input of the ``concat_gen`` function.

    Raises
    ------
    ValueError
        If there is an inconsistency in the packages.

    Examples
    --------
    >>> from raisin.serialization.iter_tools import deconcat_gen
    >>> list(deconcat_gen(pack=b'\x81a\x85houla\x80\x84hihi'))
    [b'a', b'houla', b'', b'hihi']
    >>> list(deconcat_gen(gen=[b'\x81a\x85ho', b'ula\x80\x84hihi']))
    [b'a', b'houla', b'', b'hihi']
    >>>
    """
    from raisin.serialization.core import tag2size

    n_pack = 0

    while True:
        # length recovery
        n_pack += 1
        try:
            size, pack, gen = tag2size(pack=pack, gen=gen)
        except ValueError as err:
            raise ValueError(
                'error in the header of package number ' f'{n_pack}. The flag is corrupted.'
            ) from err
        except StopIteration:
            break

        # concatenation of the enough number of data
        while len(pack) < size:
            try:
                pack += next(gen)
            except StopIteration as err:
                raise ValueError(
                    f'The package number {n_pack} is incomplete.\n'
                    f'Expected length: {size}\n'
                    f'Available length: {len(pack)}'
                ) from err

        # truncation to the exact length
        yield pack[:size]
        pack = pack[size:]


def relocate(gen):
    """
    ** Exhausts the input generator, identity function. **

    Parameters
    ----------
    gen : generator
        A byte string generator.

    Returns
    -------
    generator
        A kind of copy of the generator *gen*.

    Notes
    -----
    If the cumulative size of the data transferred by the *gen* generator
    does not exceed *BUFFER_SIZE*, then the buffering is done in RAM,
    otherwise the packets are stored in a temporary file in order to preserve RAM.

    Examples
    --------
    >>> from raisin.serialization.iter_tools import relocate
    >>>
    >>> gen = iter([b'a', b'b'])
    >>> next(gen)
    b'a'
    >>> gen_copy = relocate(gen)
    >>> next(gen)
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
    StopIteration
    >>> next(gen_copy)
    b'b'
    >>>
    """
    # the garbage collector is responsible for closing the file
    file = tempfile.SpooledTemporaryFile(max_size=BUFFER_SIZE, mode='w+b')
    for pack in concat_gen(gen):
        file.write(pack)
    file.seek(0)
    pack, gen = data2gen(file)
    return deconcat_gen(pack=pack, gen=gen)


def resize(nbr, *, pack=b'', gen=(lambda: (yield from []))()):
    """
    ** Allows to extract a precise number of bytes. **

    Parameters
    ----------
    nbr : int
        The precise number of bytes to extract.
    pack : bytes
        The first element of the generator.
    gen : generator
        The rest of the packages to be rearranged.

    Returns
    -------
    pack : bytes
        A string that contains exactly *nbr* bytes.
    gen : generator
        The generator provides in input possibly iterated or on the contrary,
        possibly lengthened if the input *pack* is bigger than the output one.

    Raises
    ------
    RuntimeError
        If there is not enough data to reach the requested length.

    Examples
    --------
    >>> from raisin.serialization.iter_tools import resize
    >>> s = lambda out: (out[0], b''.join(out[1]))
    >>> s(resize(0))
    (b'', b'')
    >>> s(resize(0, pack=b'a'))
    (b'', b'a')
    >>> s(resize(0, gen=[b'a']))
    (b'', b'a')
    >>> s(resize(0, pack=b'a', gen=[b'b']))
    (b'', b'ab')
    >>> s(resize(1, pack=b'abc', gen=[b'de', b'fgh']))
    (b'a', b'bcdefgh')
    >>> s(resize(4, pack=b'abc', gen=[b'de', b'fgh']))
    (b'abcd', b'efgh')
    >>>
    """
    iter_gen = iter(gen)
    while len(pack) < nbr:
        try:
            pack += next(iter_gen)
        except StopIteration as err:
            raise RuntimeError(
                f'only {len(pack)} bytes are available while {nbr} are requested'
            ) from err
    if len(pack) > nbr:
        ext_gen = itertools.chain([pack[nbr:]], iter_gen)
        pack = pack[:nbr]
        return pack, ext_gen
    return pack, iter_gen


def to_gen(*, pack=_Null, gen=(lambda: (yield from []))(), size=BUFFER_SIZE):
    r"""
    ** Normalize the size of the packages. **

    Group the ceded packets by 'gen' in order to cede packets
    of normalized length.

    Parameters
    ----------
    pack : bytes
        The first pack.
    gen : generator
        The generator that assigns the sequence of pack.
    size : int or None, default=BUFFER_SIZE
        Size of yieled packages (if int).
        (if None), just make a generator whose first element is 'pack'.
        All the packets transferred are then identical to those of
        the input generator.

    Yields
    ------
    bytes
        A packet of size *size*, created by concatenating the input data.

    Notes
    -----
    There is no verification on the entries.

    Examples
    --------
    >>> import random
    >>> from raisin.serialization.iter_tools import to_gen
    >>>
    >>> random.seed(0)
    >>>
    >>> l = [b'\x00'*random.randint(0, 1000) for i in range(4)]
    >>> [len(e) for e in l]
    [864, 394, 776, 911]
    >>>
    >>> [len(e) for e in to_gen(gen=l, size=500)]
    [500, 500, 500, 500, 500, 445]
    >>> [len(e) for e in to_gen(pack=b'', gen=l, size=500)]
    [500, 500, 500, 500, 500, 445]
    >>> [len(e) for e in to_gen(pack=b'\x00'*55, gen=l, size=500)]
    [500, 500, 500, 500, 500, 500]
    >>> [len(e) for e in to_gen(pack=b'\x00'*56, gen=l, size=500)]
    [500, 500, 500, 500, 500, 500, 1]
    >>>
    >>> [len(e) for e in to_gen(gen=l, size=None)]
    [864, 394, 776, 911]
    >>> [len(e) for e in to_gen(pack=b'\x00'*55, gen=l, size=None)]
    [55, 864, 394, 776, 911]
    >>> [len(e) for e in to_gen(pack=b'', gen=l, size=None)]
    [0, 864, 394, 776, 911]
    >>>
    """
    if size is None:
        if pack is not _Null:
            yield pack
        yield from gen
    else:
        if pack is _Null:
            pack = b''
        else:
            while len(pack) > size:
                yield pack[:size]
                pack = pack[size:]
        for data in gen:
            pack += data
            while len(pack) > size:
                yield pack[:size]
                pack = pack[size:]
        yield pack



#!/usr/bin/env python3

"""
** Lists the atomic functions that serialize each case separately. **
---------------------------------------------------------------------

The main serialization function ``raisin.serialization.serialize``,
looks at the nature of the object to serialize.
It applies the appropriate function by picking it up here.
It is also here that the deserialization function
``raisin.serialization.deserialize`` will draw the right function.
"""

import inspect

from raisin.serialization.atoms import (small, dumps,
    raisin_class, iterator, source_code)


__all__ = ['SERIALIZE_TABLE', 'DESERIALIZE_TABLE']
MODULES = (small, dumps, raisin_class, iterator, source_code)


SERIALIZE_TABLE = {
    func_name[10:]: func
    for module in MODULES
    for func_name, func in inspect.getmembers(module)
    if func_name.startswith('serialize_')
}

DESERIALIZE_TABLE = {
    func_name[12:]: func
    for module in MODULES
    for func_name, func in inspect.getmembers(module)
    if func_name.startswith('deserialize_')
}



#!/usr/bin/env python3

"""
** core of serialization and deserialization. **
------------------------------------------------

This is the entry point for the function that manages
the analysis of objects and serialized chains.
This is where the link is made between the single-object
serialization/deserialization functions and the global serialization
management which is able to manage different types of objects.
"""

import os

from raisin.serialization.constants import (ALPHABET, ALPHABET2INDEX,
    BUFFER_SIZE, BYTES2HEADER, HEADERLEN, N_BYTES, N_SYMB)
from raisin.serialization.iter_tools import anticipate, to_gen
from raisin.serialization.atoms import SERIALIZE_TABLE, DESERIALIZE_TABLE


def serialize(obj, *, compresslevel, copy_file, psw, authenticity, paralleling_rate):
    """
    ** Allows to add cryptographic and compression post-processes. **

    Parameters
    ----------
    They are detailed in ``raisin.serialization.serialize``.

    Yields
    ------
        Packets of bytes.

    Notes
    -----
    This function considers that the inputs are corect.
    There is no verification done on the inputs.
    In case your program controls the inputs, you should use this function.
    Otherwise, use the ``raisin.serialization.serialize`` function.
    """
    kind = obj.__class__.__name__.lower()
    if kind in SERIALIZE_TABLE:
        ser_obj = SERIALIZE_TABLE[kind](
            obj=obj,
            compact=bool(compresslevel),
            copy_file=copy_file,
            paralleling_rate=paralleling_rate)
    else:
        raise TypeError(f"object of type '{kind}' is not raisin serializable")

    # compression and encryption
    if compresslevel > 1:
        raise NotImplementedError('pas possible de compresser pour le moment')
    if authenticity:
        raise NotImplementedError('pas de hash dispo')
    if psw is not None:
        raise NotImplementedError('pas de chiffrage')
    yield from ser_obj


def deserialize(pack, gen, *, psw, paralleling_rate):
    """
    ** Allows to call the right functions to deserialize the data stream. **

    Parameters
    ----------
    pack : bytes
        The first part of the serialized object.
    gen : generator
        The continuation of the serialized object in the form of a byte string generator.

    Returns
    -------
    object
        The deserialized object.

    Other Parameters
    ----------------
    They are detailed in ``raisin.serialization.deserialize``.

    Notes
    -----
    This function considers that the inputs are corect.
    There is no verification done on the inputs.
    In case your program controls the inputs, you should use this function.
    Otherwise, use the ``raisin.serialization.deserialize`` function.
    The normalization of the inputs is done with ``raisin.serialization.iter_tools.to_gen``.
    """
    head, pack, gen = get_header(pack=pack, gen=gen)
    if head in DESERIALIZE_TABLE:
        return DESERIALIZE_TABLE[head](
            pack=pack, gen=gen, psw=psw, paralleling_rate=paralleling_rate
        )
    raise NotImplementedError(f"deserialization of '{head}' is not implemented.")


def bytes2str(gen):
    r"""
    ** Convert a batch of bytes to an ascii string. **

    Does not make any verifications on the inputs.
    Groups *N_BYTES* bytes in *N_SYMB* characters or *3* bytes in *4*
    characters depending on what allows to compact the result.
    Bijection of the ``str2bytes`` function.

    Parameters
    ----------
    gen : generator
        Generator coming directly from ``raisin.serialization.serialize``.
        It must not be empty or exhausted.
        Bijection of the ``str2bytes`` function.

    Returns
    -------
    str
        A printable ascii character string that allows to reconstitute the byte string.

    Notes
    -----
    adds a byte in the header that allows to know if the grouping is by small or large packets.

    Examples
    --------
    >>> from raisin.serialization.core import bytes2str
    >>> bytes2str([b'toto'])
    '0Dg90aabVb'
    >>> bytes2str([b'\x00'*10]*4)
    '1aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaO'
    >>>
    """

    def int2str(nbr, n_symb, n_car):
        """
        ** Convert an integer to a string. **

        nbr <= n_car**n_symb

        Parameters
        ----------
        nbr : int
            The integer to be converted.
        n_symb : int
            The size of the output message.
        n_car : int
            The size of the alphabet to consider

        Returns
        -------
        str
            A string containing *n_symb* characters chosen
            from the alphabet *ALPHABET*.
        """
        string = ''
        while nbr:
            nbr, rest = divmod(nbr, n_car)
            string = ALPHABET[rest] + string
        string = ALPHABET[0] * (n_symb - len(string)) + string
        return string

    # preparation (n_car**n_symb >= n2**n_bytes)
    pack = b''.join(gen)
    if len(pack) < N_BYTES * 3 / 4:
        out = '0'
        n_car = 64
        n_bytes = 3
        n_symb = 4
    else:
        out = '1'
        n_car = len(ALPHABET)
        n_bytes = N_BYTES
        n_symb = N_SYMB

    # convertion
    for is_end, pack in anticipate(to_gen(pack=pack, size=n_bytes)):
        out += int2str(int.from_bytes(pack, byteorder='big', signed=False), n_symb, n_car)
        if is_end and len(pack) < n_bytes:
            return out + ALPHABET[len(pack)]
    return out


def str2bytes(string):
    """
    ** Convert an ascii string to a batch of bytes. **

    Does not make any verifications on the inputs.
    Bijection of the ``bytes2str`` function.

    Parameters
    ----------
    string : str
        A string from the str2bytes function.

    Returns
    -------
    bytes
        The bytes just before they are transformed into str.

    Raises
    ------
    ValueError
        If the header does not match.

    Examples
    --------
    >>> from raisin.serialization.core import str2bytes
    >>> str2bytes('0Dg90aabVb')
    b'toto'
    >>>
    """

    def alph_to_pack(phrases, n_car, n_bytes):
        """
        ** Convert a string tensor to a string of bytes. **
        """
        return b''.join(
            (
                sum(
                    ALPHABET2INDEX[symb] * n_car ** symb_po
                    for symb_po, symb in enumerate(reversed(phrase))
                ).to_bytes(n_bytes, byteorder='big', signed=False)
                for phrase in phrases
            )
        )

    # preparation (n_car**n_symb >= n2**n_bytes)
    head, string = string[0], string[1:]
    if head == '0':
        n_car = 64
        n_bytes = 3
        n_symb = 4
    elif head == '1':
        n_car = len(ALPHABET)
        n_bytes = N_BYTES
        n_symb = N_SYMB
    else:
        raise ValueError('The header can only be ' f"'0' or '1'. Not '{head}'.")

    # desencapsulation
    data = b''
    phrases = []
    bloc = BUFFER_SIZE // n_bytes
    for i, (is_end, phrase) in enumerate(anticipate(to_gen(pack=string, size=n_symb))):
        if is_end and len(phrase) < n_symb:
            data += alph_to_pack(phrases, n_car, n_bytes)
            data = (
                data[:-n_bytes]
                + data[
                    -sum(
                        ALPHABET2INDEX[symb] * n_car ** symb_pos
                        for symb_pos, symb in enumerate(reversed(phrase))
                    ) :
                ]
            )
            return data
        phrases.append(phrase)
        if not i % bloc:
            data += alph_to_pack(phrases, n_car, n_bytes)
            phrases = []
    data += alph_to_pack(phrases, n_car, n_bytes)
    return data


def is_jsonisable(obj, copy_file):
    """
    ** Checks if the object can be serialized with json. **

    Examples
    --------
    >>> from raisin.serialization.core import is_jsonisable
    >>> copy_file = True
    >>> is_jsonisable(0, copy_file)
    True
    >>> is_jsonisable(.0, copy_file)
    True
    >>> is_jsonisable(True, copy_file)
    True
    >>> is_jsonisable(None, copy_file)
    True
    >>> is_jsonisable('a string', copy_file)
    True
    >>> is_jsonisable(b'a bytes', copy_file)
    False
    >>> is_jsonisable(('a tuple',), copy_file)
    False
    >>> is_jsonisable([1, 2, 3], copy_file)
    True
    >>> is_jsonisable([1, b'', 3], copy_file)
    False
    >>> is_jsonisable({1: 'a'}, copy_file)
    True
    >>> is_jsonisable({1: b'a'}, copy_file)
    False
    >>>
    """
    kind = obj.__class__.__name__
    if kind in {'int', 'float', 'bool', 'NoneType'}:
        return True
    if kind == 'str' and ((not copy_file) or (len(obj) < 32767 and not os.path.isfile(obj))):
        return True
    if kind == 'list':
        return all(is_jsonisable(elem, copy_file) for elem in obj)
    if kind == 'dict':
        return all(
            is_jsonisable(key, copy_file) and is_jsonisable(value, copy_file)
            for key, value in obj.items()
        )
    return False


def size2tag(size):
    r"""
    ** Creates a flag that allows to group *size* bytes. **

    Parameters
    ----------
    size : int
        The number that can be encoded in the flag. >= 0

    Returns
    -------
    bytes
        The flags of variable size according to the size
        of the message to encode.

    Notes
    -----
    Can be very useful to record the size of a frame and be able
    to reconstruct it using the ``tag2size`` function.

    Examples
    --------
    >>> from raisin.serialization.core import size2tag
    >>> p_bin = lambda flag: ', '.join(bin(o) for o in flag)
    >>>
    >>> size2tag(0)
    b'\x80'
    >>> p_bin(_)
    '0b10000000'
    >>> size2tag(1)
    b'\x81'
    >>> p_bin(_)
    '0b10000001'
    >>> size2tag(127)
    b'\xff'
    >>> p_bin(_)
    '0b11111111'
    >>> size2tag(128)
    b'\x01\x80'
    >>> p_bin(_)
    '0b1, 0b10000000'
    >>> size2tag(2**(7*3))
    b'\x01\x00\x00\x80'
    >>> p_bin(_)
    '0b1, 0b0, 0b0, 0b10000000'
    >>>
    """
    if size == 0:
        return b'\x80'
    flags_list = []
    while size:
        size, rest = divmod(size, 2 ** 7)
        flags_list.append(rest)
    flags_list.reverse()
    flags_list[-1] += 1 << 7
    return bytes(flags_list)


def tag2size(*, pack=b'', gen=(lambda: (yield from []))()):
    r"""
    ** Allows to find the number encoded in a flag. **

    Parameters
    ----------
    pack : bytes
        The first package of the *gen* generator.
        Must contain the beginning of the flag.
    gen : generator
        The generator that contains the sequence of bytes
        if they are not all specified in *pack*.

    Returns
    -------
    size : int
        The length represented by the flag.
    pack : bytes
        The entry package without the flag.
    gen : generator
        The *gen* generator possibly slightly iterated.

    Raises
    ------
    StopIteration
        If the flag is empty.
    RuntimeError
        If the flag is incomplete.

    Notes
    -----
    This is the bijection of the ``tag2size`` function.

    Examples
    --------
    >>> from raisin.serialization.core import tag2size
    >>>
    >>> tag2size(pack=b'\x80')[0]
    0
    >>> tag2size(pack=b'\x81')[0]
    1
    >>> tag2size(pack=b'\xff')[0]
    127
    >>> tag2size(pack=b'\x01\x80')[0]
    128
    >>> tag2size(pack=b'\x01\x00\x00\x80')[0]
    2097152
    >>>
    """
    gen = iter(gen)
    size = 0
    while True:
        if pack:
            octet, pack = pack[0], pack[1:]
        else:
            while not pack:
                try:
                    pack = next(gen)
                except StopIteration as err:
                    if size:
                        raise RuntimeError('the flag is not complete') from err
                    raise StopIteration('there are no data') from err
            octet, pack = pack[0], pack[1:]
        if octet & 0b10000000:  # if we reach the end of the flag
            size = (size << 7) + (octet - 0b10000000)
            return size, pack, gen
        size = (size << 7) + octet


def get_header(*, pack=b'', gen=(lambda: (yield from []))()):
    r"""
    ** Retrieves the name of the header of a serialized object. **

    Leve une ValueError si l'entete n'est pas correcte.

    Parameters
    ----------
    pack : bytes
        The first package of the *gen* generator.
        Must contain the beginning of the header.
    gen : generator
        The generator that contains the sequence of bytes
        if they are not all specified in *pack*.

    Returns
    -------
    header : str
        The name of the header. This is a dictionary key *HEADER*.
    pack : bytes
        The entry package without the flag.
    gen : generator
        The *gen* generator possibly slightly iterated.

    Raises
    ------
    ValueError
        In case the header is not valid or incomplete.

    Examples
    --------
    >>> from raisin.serialization.core import get_header
    >>>
    >>> get_header(pack=b'/pickle/')[0]
    'pickle'
    >>> get_header(pack=b'\x00')[0]
    'pickle'
    >>> get_header(pack=b'/pickle/and ...')[0]
    'pickle'
    >>>
    """
    gen = iter(gen)
    bytes_header = b''
    while True:
        while not pack:
            try:
                pack = next(gen)
            except StopIteration as err:
                raise ValueError(f'the header is too short: {bytes_header}') from err
        bytes_header += pack[:1]
        pack = pack[1:]
        try:
            return BYTES2HEADER[bytes_header], pack, gen
        except KeyError as err:
            if not HEADERLEN[len(bytes_header)]:
                raise ValueError(
                    f'the bytes sequence {bytes_header} does not ' 'correspond to any known header.'
                ) from err



#!/usr/bin/env python3

"""
** Forms the packages. **
-------------------------

Allows to generate packages.
"""

import queue
import threading

from raisin.encapsulation.packaging import Argument, Func, Task
from raisin.communication.client import Client


__pdoc__ = {'TasksSharer.__enter__': True, 'TasksSharer.__exit__': True}


def make_tasks(functions, arguments):
    """
    ** Generates the tasks to be performed. **

    Parameters
    ----------
    functions : iterable
        Generator that assigns each function associated with each task.
        In the case of map, this iterator must yield the same function at each iteration.
    arguments : iterable
        Generator that assigns the set of arguments for each task. At each iteration
        the arguments must be provided via a tuple of size *n*, with *n* the number
        of arguments that the function associated with this task takes.

    Yields
    ------
    new_packages : list
        The list of packets to send for the proper functioning of the task.
        This list contains packets of type ``raisin.encapsulation.packaging.Func``
        or ``raisin.encapsulation.packaging.Argument``.
        This list does not contain packages that have already been assigned
        to a previous task, even if they are also needed for the current task.
    task : ``raisin.encapsulation.packaging.Task``
        The task to execute, which contains the address of the function and its arguments.

    Notes
    -----
    - There is no verification on the inputs, they must be verified by higher level functions.
    - For functions, the detection of redundancy is done from the address of the memory pointer.

    Examples
    --------
    >>> from itertools import cycle
    >>> from raisin.encapsulation.manager import make_tasks
    >>>
    >>> def f(x, y):
    ...     return x**2 + y**2
    ...
    >>> def pprint(job):
    ...    for new_packages, task in job:
    ...        print(f'new: {new_packages}, task: {task}')
    ...
    >>> pprint(make_tasks(cycle((f,)), [(1, 2), (2, 3)]))
    new: [Argument(1), Argument(2), Func(f)], task: Task(3689584644591251701, [12402937695876512259, 15584102817361292984])
    new: [Argument(3)], task: Task(3689584644591251701, [15584102817361292984, 15229981472126021560])
    >>>
    """

    def check_and_add(set_, element):
        if element in set_:
            return False
        set_.add(element)
        return True

    pointers2hashes = {}  # To each object address, associates its hash.
    args_hashes = set()  # The hash value of each argument.
    for func, args in zip(functions, arguments):
        new_args = [(a, Argument(a)) for a in args if id(a) not in pointers2hashes]
        pointers2hashes.update({id(p): a.__hash__() for p, a in new_args})
        new_args = [a for _, a in new_args if check_and_add(args_hashes, a.__hash__())]

        new_func = Func(func) if id(func) not in pointers2hashes else None
        if new_func is not None:
            pointers2hashes[id(func)] = new_func.__hash__()

        new_packages = new_args if new_func is None else new_args + [new_func]
        task = Task(
            func_hash=pointers2hashes[id(func)], arg_hashes=[pointers2hashes[id(a)] for a in args]
        )
        yield new_packages, task


class TasksSharer(threading.Thread):
    """
    ** Communicate packets with the local server. **

    Attributes
    ----------
    res_queue : Queue
        The FIFO queue which contains the result packets of type
        ``raisin.encapsulation.packaging.Result` in the order they arrive.

    Examples
    --------
    >>> from raisin.encapsulation.manager import make_tasks, TasksSharer
    >>>
    >>> def f(x): return x**2
    ...
    >>> tasks_generator = make_tasks((f,), ([2],))
    >>> with TasksSharer(tasks_generator) as tasks_sharer:
    ...     tasks_sharer.start()
    ...     tasks_sharer.join()
    ...
    >>>
    """

    def __init__(self, tasks_iterator):
        """
        Parameters
        ----------
        tasks_iterator : iterator
            Gives for each task, the list of packages of type
            ``raisin.encapsulation.packaging.Func``
            or ``raisin.encapsulation.packaging.Argument`` and also the task to execute
            (of type ``raisin.encapsulation.packaging.Task``). For example, the generator
            ``raisin.encapsulation.manager.make_tasks`` can do the job.

        Notes
        -----
        - There is no verification on the inputs, they must be verified by higher level functions.
        - You have to start the thread with the *start()* method to establish
            asynchronous communication with the server.
        """
        threading.Thread.__init__(self)
        self._tasks_iterator = tasks_iterator
        self.res_queue = queue.Queue()

    def __enter__(self):
        """
        ** Allows to cut the connection in case of error. **

        Instead of quitting the server without warning, this context
        manager allows you to quit the dialog cleanly.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        ** Stops and closes the dialog with the local server. **

        Goes together with ``TasksSharer.__enter__``.
        This is where you break the connection with the server.
        """
        return None

    def run(self):
        """
        ** Represents the thread activity. **

        This method is the first method that will be executed when the thread is launched.
        It must not be called directly. It is the call of the 'start' method which
        takes care in the background to launch this method.
        """
        # try:
        #     with Client(None, 20001) as client:
        #         client.start()
        #         for new_packages, task in self._tasks_iterator:
        #             for new_package in new_packages:
        #                 client.send_obj(new_package)
        #             client.send_obj(task)
        # except ConnectionError as err:
        #     raise ConnectionError('please launch a listening server on port 20001') from err



#!/usr/bin/env python3

"""
** Processes requests. **
-------------------------

Whether it is a client or a server that asks for something, it doesn't change much.
In all cases it is TCP sockets that request and render services.
That's why once the communication is established,
clients and servers use the same function to communicate.
"""

import threading


class Handler(threading.Thread):
    """
    ** Helps a socket to communicate. **

    Attributes
    ----------
    abstract_conn : AbstractConn
        The abstract connection that allows communication.
    """

    def __init__(self, abstract_conn):
        """
        Parameters
        ----------
        abstract_conn : AbstractConn
            An entity able to communicate.
        """
        threading.Thread.__init__(self)
        self.abstract_conn = abstract_conn

    def run(self):
        """
        ** Wait for the requests to answer them. **

        This method must be launched asynchronously by invoking the *start* method.
        It listens for the arrival of a request through the 'abstract_conn' attribute.
        As soon as a request arrives, it is processed. Once the request is processed,
        this method starts listening for the next request.
        """
        while True:
            try:
                request = self.abstract_conn.recv_obj()
            except ConnectionError:
                break
            raise NotImplementedError(f'impossible de traiter {request}, c est pas code')

    def handler_close(self):
        """
        ** Clean up the connection. **
        """
        self.abstract_conn.close()



#!/usr/bin/env python3

"""
** Unitary creation of packets. **
----------------------------------

This is where there are the basic classes that represent the fundamental
packets for transferring data across the network.
"""

import abc
import hashlib

from raisin.serialization import deserialize, serialize

__pdoc__ = {
    'Package.__repr__': True,
    'Package.__hash__': True,
    'Package.__eq__': True,
    'Package.__getstate__': True,
    'Package.__setstate__': True}


class Package:
    """
    ** Base class to represent a package. **
    """

    def __repr__(self):
        """
        ** Improve representation. **

        Returns an evaluable representation in a particular context
        of the dynamic instance of an object inherited from this class.

        Examples
        --------
        >>> from raisin.encapsulation.packaging import Package
        >>> Package()
        Package()
        >>>
        """
        return f'{self.__class__.__name__}()'

    def __hash__(self):
        """
        ** Compute a real unique hash. **

        The hash calculation is based on the content of the object,
        not on the addresses in memory.

        Examples
        --------
        >>> from raisin.encapsulation.packaging import Package
        >>> {Package(), Package()}
        {Package()}
        >>>
        """
        if self.__class__.__name__ == 'Package':
            return 0
        return int(
            hashlib.sha1(b''.join(serialize(self.__getstate__(), compresslevel=1))).hexdigest(), base=16
        ) % (2**64) # TODO : Make a real hash function!

    def __eq__(self, other):
        """
        ** Compare 2 objects based on the hash. **

        2 objects are equal if they are exactly
        the same type and have the same hash value.
        This function is essential to be able to apply
        hashable collections on ``Package`` instance.

        Examples
        --------
        >>> from raisin.encapsulation.packaging import Package
        >>> Package() == Package()
        True
        >>>
        """
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        return hash(self) == hash(other)

    @abc.abstractmethod
    def __getstate__(self, as_iter=False, **kwds):
        """
        ** Help for the serialization. **

        Must be overwritten.

        Parameters
        ----------
        as_iter : boolean
            If True, returns a generator rather than binary.
            This can be useful to preserve RAM.
        **kwds : dict
            Arguments passed to the function ``raisin.serialization.serialize``.

        Returns
        -------
        state : bytes or iterable
            Sufficient information to reconstruct the object.
        """
        raise NotImplementedError('this abstract method must be overwritten')

    @abc.abstractmethod
    def __setstate__(self, state, **kwds):
        """
        ** Allows you to reconstruct the object. **

        Must be overwritten.
        This method replaces *__init__*. It is for example used by *pickle*.

        Parameters
        ----------
        state : bytes or iterable
            The state of the object is the value returned
            by the method ``Argument.__getstate__``.
        **kwds : dict
            Arguments passed to the function ``raisin.serialization.deserialize``.
        """
        raise NotImplementedError('this abstract method must be overwritten')


class Argument(Package):
    """
    ** Represents an argument for a function. **
    """

    def __init__(self, arg):
        self._arg = arg

    def get_value(self):
        """
        ** Accessor to retrieve the value of the argument. **

        Example
        -------
        >>> from raisin.encapsulation.packaging import Argument
        >>> Argument(0)
        Argument(0)
        >>> _.get_value()
        0
        >>>
        """
        return self._arg

    def __repr__(self):
        """
        ** Improve representation. **

        Examples
        --------
        >>> from raisin.encapsulation.packaging import Argument
        >>> Argument(0)
        Argument(0)
        >>> a = 1
        >>> Argument(a)
        Argument(1)
        >>> Argument('toto')
        Argument('toto')
        >>> Argument(10**100)
        Argument(<arg>)
        >>>
        """
        arg_name = repr(self._arg)
        if len(arg_name) > 100:
            arg_name = '<arg>'
        return f'Argument({arg_name})'

    def __getstate__(self, as_iter=False, **kwds):
        """
        ** Implementation of ``Package.__getstate__``. **

        Examples
        --------
        >>> from pickle import dumps
        >>> from raisin.encapsulation.packaging import Argument
        >>> Argument(0)
        Argument(0)
        >>> type(dumps(_))
        <class 'bytes'>
        >>>
        """
        if as_iter is False:
            return b''.join(self.__getstate__(as_iter=True, **kwds))
        return serialize(self.get_value(), **kwds)

    def __setstate__(self, state, **kwds):
        """
        ** Implementation of ``Package.__setstate__``. **

        Examples
        --------
        >>> from pickle import dumps, loads
        >>> from raisin.encapsulation.packaging import Argument
        >>> Argument(0)
        Argument(0)
        >>> loads(dumps(_))
        Argument(0)
        >>>
        """
        self._arg = deserialize(state, **kwds)


class Func(Package):
    """
    ** Encapsulate a simple function. **
    """

    def __init__(self, func):
        self._func = func

    def __repr__(self):
        """
        ** Improve representation. **

        Examples
        --------
        >>> from raisin.encapsulation.packaging import Func
        >>> def f(): pass
        >>> Func(f)
        Func(f)
        >>> g = f
        >>> Func(g)
        Func(f)
        >>> Func(lambda x : None)
        Func(<lambda>)
        >>>
        """
        try:
            return f'Func({self._func.__name__})'
        except AttributeError:
            return 'Func(<lambda>)'

    def __hash__(self):
        """
        ** Implementation of ``Package.__hash__``. **
        """
        return int(
            hashlib.sha1(self._func.__name__.encode()).hexdigest(), base=16
        ) % (2**64) # TODO : Make a real hash function!

    def __getstate__(self, as_iter=False, **kwds):
        """
        ** Implementation of ``Package.__getstate__``. **

        Examples
        --------
        >>> from pickle import dumps
        >>> from raisin.encapsulation.packaging import Func
        >>> Func(lambda x: x**2)
        Func(<lambda>)
        >>> type(dumps(_))
        <class 'bytes'>
        >>>
        """
        if as_iter is False:
            return b''.join(self.__getstate__(as_iter=True, **kwds))
        return serialize(self._func, **kwds)

    def __setstate__(self, state, **kwds):
        """
        ** Implementation of ``Package.__setstate__``. **

        Examples
        --------
        >>> from pickle import dumps, loads
        >>> from raisin.encapsulation.packaging import Func
        >>> Func(lambda x: x**2)
        Func(<lambda>)
        >>> loads(dumps(_))
        Func(<lambda>)
        >>>
        """
        self._func = deserialize(state, **kwds)


class Result(Package):
    """
    ** Represents the result of a job. **
    """

    def __init__(self):
        pass

    def __getstate__(self, as_iter=False, **_):
        """
        ** Implementation of ``Package.__getstate__``. **
        """
        if as_iter is False:
            return b''
        return (lambda: (yield b''))()

    def __setstate__(self, state, **_):
        """
        ** Implementation of ``Package.__setstate__``. **
        """


class Task(Package):
    """
    ** Complete block to perform a calculation. **
    """

    def __init__(self, func_hash, arg_hashes):
        self.func_hash = func_hash
        self.arg_hashes = arg_hashes

    def __repr__(self):
        """
        ** Improve representation. **

        Examples
        --------
        >>> from raisin.encapsulation.packaging import Argument, Func, Task
        >>> func = Func(lambda x: x**2)
        >>> arg1 = Argument(0)
        >>> arg2 = Argument(1)
        >>> Task(func.__hash__(), (arg1.__hash__(), arg2.__hash__()))
        Task(9134397370399107984, (8576677656317280732, 12402937695876512259))
        >>>
        """
        return f'Task({self.func_hash}, {self.arg_hashes})'

    def __getstate__(self, as_iter=False, **kwds):
        """
        ** Implementation of ``Package.__getstate__``. **

        Examples
        --------
        >>> from pickle import dumps
        >>> from raisin.encapsulation.packaging import Argument, Func, Task
        >>> func = Func(lambda x: x**2)
        >>> arg1 = Argument(0)
        >>> arg2 = Argument(1)
        >>> Task(func.__hash__(), (arg1.__hash__(), arg2.__hash__()))
        Task(9134397370399107984, (8576677656317280732, 12402937695876512259))
        >>> type(dumps(_))
        <class 'bytes'>
        >>>
        """
        if as_iter is False:
            return b''.join(self.__getstate__(as_iter=True, **kwds))
        return serialize((self.func_hash, self.arg_hashes), **kwds)

    def __setstate__(self, state, **kwds):
        """
        ** Implementation of ``Package.__setstate__``. **

        Examples
        --------
        >>> from pickle import dumps, loads
        >>> from raisin.encapsulation.packaging import Argument, Func, Task
        >>> func = Func(lambda x: x**2)
        >>> arg1 = Argument(0)
        >>> arg2 = Argument(1)
        >>> Task(func.__hash__(), (arg1.__hash__(), arg2.__hash__()))
        Task(9134397370399107984, (8576677656317280732, 12402937695876512259))
        >>> loads(dumps(_))
        Task(9134397370399107984, (8576677656317280732, 12402937695876512259))
        >>>
        """
        self.func_hash, self.arg_hashes = deserialize(state, **kwds)



#!/usr/bin/env python3

"""
** Serialize functions and classes. **

Rely heavily on the *dill* module.
"""

import dill

from raisin.serialization.constants import HEADER


def serialize_function(obj, compact, **_):
    """
    ** Serialize the functions. **
    """
    yield HEADER['function'][compact]
    yield dill.dumps(obj, recurse=True)

def deserialize_function(pack, gen, **_):
    """
    ** Deserialize the functions. **
    """
    return dill.loads(pack + b''.join(gen))



#!/usr/bin/env python3

"""
** Used to encapsulate packets before sending them. **
------------------------------------------------------

In particular, when the user wants to execute several tasks through a network,
they are split into small packets. These packets are made up of the function
and the arguments specific to each task.
"""

__all__ = ['Argument', 'Func', 'Result', 'Task']

from raisin.encapsulation.packaging import Argument, Func, Result, Task



#!/usr/bin/env python3

r"""
** Raisin: To perform cluster work easily! **
---------------------------------------------

The main aim of project *raisin* is to **share physical resources** of your laptop
with a community.
In counterpart, you can **benefit from the community resources**.

Notes
-----
* To generate the documentation, you have to install the package *pdoc3* with *pip* for example.
    Then, you just have to type the following command to generate the html files:
    * ``pdoc3 raisin/ -c latex_math=True --force --html``
* To run the test benches, you have to install the package *pytest* with *pip* for example.
    Then you have to type the following command:
    * ``clear && pytest --doctest-modules raisin/
        && find raisin/ -regex '.*test.*\.py' -exec pytest {} \;``
* Docstrings respect the following convention:
    * ``https://numpydoc.readthedocs.io/en/latest/format.html``
* For text formatting:
    * ``black -C -S -l 100 raisin``
"""

import inspect

from raisin.serialization import deserialize, dump, dumps, load, loads, serialize


__version__ = '3.0.0'
__author__ = 'Robin RICHARD (robinechuca) <raisin@ecomail.fr>'
__license__ = 'GNU Affero General Public License v3 or later (AGPLv3+)'
__all__ = ['deserialize', 'dump', 'dumps', 'load', 'loads', 'serialize']
__pdoc__ = {
    obj: 'Alias to ``raisin.{}.{}``'.format(
        (inspect.getsourcefile(globals()[obj]).split('raisin/')[-1][:-3])
        .replace('/', '.')
        .replace('.__init__', ''),
        obj,
    )
    for obj in __all__
}
__pdoc__ = {
    **__pdoc__,
    **{
        f'{cl}.{meth}': False
        for cl in __all__
        if globals()[cl].__class__.__name__ == 'type'
        for meth in globals()[cl].__dict__
        if not meth.startswith('_')
    },
}



#!/usr/bin/env python3

"""
** Conversion from ascii to object. **
--------------------------------------
"""


def deserialize_dumps(pack, gen, psw, paralleling_rate, **_):
    """
    ** Deserialize the objects returned by the function ``raisin.serialization.dumps``. **

    Examples
    --------
    >>> from raisin.serialization.atoms.dumps import deserialize_dumps
    >>> deserialize_dumps(pack=b'0aWDBam0vc', gen=[], psw=None, paralleling_rate=0)
    123456789
    >>>
    """
    from raisin.serialization.core import deserialize, str2bytes

    return deserialize(
        pack=str2bytes((pack + b''.join(gen)).decode(encoding='ascii')),
        gen=[],
        psw=psw,
        paralleling_rate=paralleling_rate,
    )



#!/usr/bin/env python3

"""
** Allows to serialize and deserialize generators. **

More particularly, it allows to serialize iterators,
often containers like (*list*, *tuple*, *set*...).
On the other hand, deserialization is presented in the form of a generator.
It will be necessary to recast in the right type.
"""

from raisin.serialization.constants import HEADER
from raisin.serialization.iter_tools import anticipate


def _ser_generator(obj, compact, copy_file, paralleling_rate):
    r"""
    ** Serialize the content of any iterable object. **

    Don't give in to the urge!
    This function is meant to be used for serialization of other containers.

    Examples
    --------
    >>> from raisin.serialization.atoms.iterator import _ser_generator
    >>> s = lambda obj: b''.join(
    ...     _ser_generator(obj, compact=True, copy_file=False, paralleling_rate=0))
    >>> s([])
    b''
    >>> s([None])
    b'\x82\x01\x13'
    >>> s([None, None])
    b'\x82\x01\x13\x82\x01\x13'
    >>>
    """
    from raisin.serialization.core import serialize, size2tag
    try:
        yield from (
            size2tag(1 + len(pack)) + bytes([is_end]) + pack
            for item in obj
            for is_end, pack in anticipate(
                serialize(
                    item,
                    compresslevel=compact,
                    copy_file=copy_file,
                    psw=None,
                    authenticity=False,
                    paralleling_rate=paralleling_rate,
                )
            )
        )
    except RuntimeError:  # if the generator is empty
        pass


def serialize_generator(obj, compact, copy_file, paralleling_rate):
    r"""
    ** Serialize the generators. **

    Be careful, iter on the object, in the case of a generator for example,
    it will be exhausted after the call of this function.

    Examples
    --------
    >>> from raisin.serialization.atoms.iterator import serialize_generator
    >>> s = lambda obj: b''.join(
    ...     serialize_generator(obj, compact=True, copy_file=False, paralleling_rate=0))
    >>> s([])
    b'\x18'
    >>> s([None])
    b'\x18\x82\x01\x13'
    >>> s([None, None])
    b'\x18\x82\x01\x13\x82\x01\x13'
    >>>
    """
    yield HEADER['generator'][compact]
    yield from _ser_generator(obj, compact, copy_file, paralleling_rate)


def deserialize_generator(pack, gen, psw, paralleling_rate):
    r"""
    ** Gives up the elements of the serialized iterator. **

    Yields
    ------
    The deserialized elements are equivalent to those present
    in the initial iterable, before it was serialized.

    Raises
    ------
    ValueError
        If there is an inconsistency in the data.
    RuntimeError
        If there is not enough data available.

    Examples
    --------
    >>> from raisin.serialization.atoms.iterator import deserialize_generator
    >>> list(deserialize_generator(b'', [], None, 0))
    []
    >>> list(deserialize_generator(b'\x82\x01\x13', [], None, 0))
    [None]
    >>> list(deserialize_generator(b'\x82\x01\x13\x82\x01\x13', [], None, 0))
    [None, None]
    >>>
    """
    from raisin.serialization.core import deserialize, tag2size
    from raisin.serialization.iter_tools import resize

    class IsEnd(Exception):
        """ raising to announce the end """

    class SubItems:
        """help for iter into an over iterator"""

        def __init__(self, pack, gen):
            self.pack = pack
            self.gen = iter(gen)

        def item_raw(self, item_range):
            """give away the raw data of an item"""
            subitem_range = 0
            while True:
                try:
                    size, self.pack, self.gen = tag2size(pack=self.pack, gen=self.gen)
                except StopIteration as err:
                    raise IsEnd from err
                except RuntimeError as err:
                    raise RuntimeError(
                        f'the header of subpackage n {subitem_range} '
                        f'of element n {item_range} is incomplete.'
                    ) from err

                while len(self.pack) < size:
                    try:
                        self.pack += next(self.gen)
                    except StopIteration as err:
                        raise RuntimeError(
                            f'we expect subpackage n {subitem_range} element n {item_range} to be '
                            f'serialized with {size} bytes, but only {len(self.pack)} are available'
                        ) from err
                is_end, self.pack = self.pack[0], self.pack[1:]
                if is_end > 1:
                    raise ValueError(
                        f'the byte which designates if the subpackage n {subitem_range} of element '
                        f'n {item_range} is the last one is a boolean which can only take '
                        f'the values 0 or 1, and it is {is_end}'
                    )

                data, self.pack = self.pack[:size-1], self.pack[size-1:]
                yield data
                if is_end:
                    break
                subitem_range += 1

        def get_item(self, item_range, psw, paralleling_rate):
            """deserialize the item"""
            return deserialize(
                pack=b'', gen=self.item_raw(item_range), psw=psw, paralleling_rate=paralleling_rate
            )

    sub_items = SubItems(pack=pack, gen=gen)
    item_range = 0
    while True:
        try:
            yield sub_items.get_item(item_range, psw=psw, paralleling_rate=paralleling_rate)
        except IsEnd:
            break
        item_range += 1


def serialize_tuple(obj, compact, copy_file, paralleling_rate):
    r"""
    ** Serialize tuple. **

    Examples
    --------
    >>> from raisin.serialization.atoms.iterator import serialize_tuple
    >>> s = lambda obj: b''.join(
    ...     serialize_tuple(obj, compact=True, copy_file=False, paralleling_rate=0))
    >>> s(())
    b'\x0e'
    >>> s((None,))
    b'\x0e\x82\x01\x13'
    >>> s([None, None])
    b'\x0e\x82\x01\x13\x82\x01\x13'
    >>>
    """
    yield HEADER['tuple'][compact]
    yield from _ser_generator(obj, compact, copy_file, paralleling_rate)


def deserialize_tuple(pack, gen, psw, paralleling_rate):
    r"""
    ** Deserialize tuple. **

    Examples
    --------
    >>> from raisin.serialization.atoms.iterator import deserialize_tuple
    >>> deserialize_tuple(b'', [], None, 0)
    ()
    >>> deserialize_tuple(b'\x82\x01\x13', [], None, 0)
    (None,)
    >>> deserialize_tuple(b'\x82\x01\x13\x82\x01\x13', [], None, 0)
    (None, None)
    >>>
    """
    return tuple(
        deserialize_generator(pack=pack, gen=gen, psw=psw, paralleling_rate=paralleling_rate)
    )



#!/usr/bin/env python3

"""
** Allows for extensive testing around serialization. **
--------------------------------------------------------

There is a function capable of generating many types of objects.
For each of these objects they are serialized then deserialized.
"""

import itertools
import sys

import pytest

from raisin.serialization.constants import ALPHABET, BUFFER_SIZE
from raisin.serialization.iter_tools import (anticipate, concat_gen,
    deconcat_gen, relocate, resize, to_gen)
from raisin.serialization.core import size2tag, tag2size
from raisin.serialization import dumps, loads, serialize, deserialize
from raisin.encapsulation.packaging import Argument, Func, Result, Task


# OBJECTS GENERATION

def get_int():
    """
    ** Makes some various intergers. **
    """
    yield 0
    for i in [2 ** j for j in range(1, 16)]:
        yield 2 ** i - 1
        yield 2 ** i
        yield 2 ** i + 1
        yield -(2 ** i - 1)
        yield -(2 ** i)
        yield -(2 ** i + 1)

def get_float():
    """
    ** Makes some various floats. **
    """
    yield 0.0
    yield sys.float_info.max
    yield -sys.float_info.max
    yield sys.float_info.min
    yield -sys.float_info.min
    yield 1.0

def get_simple_objects():
    """
    ** Yield some vaious simple objects. **
    """
    yield None
    yield from get_int()
    yield from get_float()


# TESTS TOOLS

def test_tools():
    """
    ** Tests the tools. **
    """
    # anticipate
    assert list(anticipate([''])) == [(True, '')]
    assert list(anticipate(['', ''])) == [(False, ''), (True, '')]
    with pytest.raises(RuntimeError):
        list(anticipate([]))

    # concat
    for gen in ([], [b''], [b'', b''], [b'a'], [b'a', b'b'], [b'', b'b'], [b'a', b'']):
        assert list(deconcat_gen(gen=concat_gen(gen))) == gen
        assert list(deconcat_gen(pack=b''.join(concat_gen(gen)))) == gen

    # relocate
    iterator = iter([])
    iterator_bis = relocate(iterator)
    with pytest.raises(StopIteration):
        next(iterator_bis)
    with pytest.raises(StopIteration):
        next(iterator)
    for gen in (
            [b'a'],
            [b'a', b'b', b'c'],
            itertools.chain([b'a'], (b'b' * BUFFER_SIZE for _ in range(100)))):
        iterator = iter(gen)
        iterator_bis = relocate(iterator)
        with pytest.raises(StopIteration):
            next(iterator)
        assert next(iterator_bis) == b'a'

    # resize
    for nbr in [0, 1, 2, 10, 100, 1000]:
        for pack_len in [0, 1, 2, 10, 100, 1000]:
            for gen_len in [0, 1, 2, 4]:
                for gen_item_len in [0, 1, 100, 500]:
                    pack = bytes(pack_len)
                    gen = (bytes(gen_item_len) for _ in range(gen_len))
                    if nbr <= pack_len + gen_len*gen_item_len:
                        pack_, gen_ = resize(nbr, pack=pack, gen=gen)
                        assert len(pack_) == nbr
                        assert len(pack_ + b''.join(gen_)) == pack_len + gen_len*gen_item_len
                    else:
                        with pytest.raises(RuntimeError):
                            resize(nbr, pack=pack, gen=gen)

    # to_gen
    for size in [1, 2, 1000, BUFFER_SIZE]:
        for gen in (
                [],
                [b''],
                [b'a'],
                [b'a' * size],
                [b'a' * (size - 1)],
                [b'a' * (size + 1)],
                [b'', b''],
                [b'a' * ((2 ** l) % (2 * size)) for l in range(100)]):
            l_in = [len(e) for e in gen]
            l_out = [len(e) for e in to_gen(gen=gen, size=size)]
            assert sum(l_in) == sum(l_out)
            assert all(s == size for s in l_out[:-1])
            assert l_out[-1] <= size
            l_out = [len(e) for e in to_gen(pack=b''.join(gen), size=size)]
            assert sum(l_in) == sum(l_out)
            assert all(s == size for s in l_out[:-1])
            assert l_out[-1] <= size

    # size & tag
    with pytest.raises(StopIteration):
        tag2size(pack=b'')
    with pytest.raises(RuntimeError):
        tag2size(pack=b'\x01\x00\x00')
    for size in [0, 1, 2, 127, 128, 129, 123456789123456789]:
        pack = size2tag(size)
        assert tag2size(pack=pack)[0] == size



# TESTS SERIALIZATION

def test_dumps():
    """
    ** Tests dumps. **
    """
    for obj in get_simple_objects():
        for compresslevel in [0, 1]:
            ser = dumps(obj, compresslevel=compresslevel)
            assert all(c in ALPHABET for c in ser)
            assert loads(ser) == obj

def test_raisin_class():
    """
    ** Tests raisin class. **
    """
    arg = Argument(0)
    func = Func(lambda x : x**2)
    task = Task(hash(func), (hash(arg),))
    result = Result()

    assert deserialize(serialize(arg)) == arg
    assert deserialize(serialize(func)) == func
    assert deserialize(serialize(task)) == task
    assert deserialize(serialize(result)) == result



#!/usr/bin/env python3

"""
** Allows to serialize and deserialize the classes of the module. **

As the source code of these classes is already known,
it is useless to serialize it too.
That's why we treat these classes separately.
"""

from raisin.serialization.constants import HEADER
from raisin.serialization.iter_tools import to_gen


def _init_class(cls, state, **kwds):
    """
    ** Constructs and instantiates a class via *__setstate__* method. **
    """
    inst = cls.__new__(cls)
    inst.__setstate__(state, **kwds)
    return inst


def serialize_argument(obj, compact, **kwds):
    r"""
    ** Serialize the ``raisin.encapsulation.packaging.Argument``. **

    Examples
    --------
    >>> from raisin.encapsulation.packaging import Argument
    >>> from raisin.serialization.atoms.raisin_class import serialize_argument
    >>> arg = Argument(0)
    >>> b''.join(serialize_argument(arg, compact=True))
    b'%\x03\x00'
    >>>
    """
    yield HEADER['argument'][compact]
    yield from obj.__getstate__(as_iter=True, compresslevel=compact, **kwds)


def deserialize_argument(pack, gen, **kwds):
    r"""
    ** Deserialize ``raisin.encapsulation.packaging.Argument``. **

    Examples
    --------
    >>> from raisin.serialization.atoms.raisin_class import deserialize_argument
    >>> deserialize_argument(pack=b'\x03\x00', gen=[])
    Argument(0)
    >>>
    """
    from raisin.encapsulation.packaging import Argument
    return _init_class(Argument, to_gen(pack=pack, gen=gen), **kwds)


def serialize_func(obj, compact, **kwds):
    r"""
    ** Serialize the ``raisin.encapsulation.packaging.Func``. **

    Examples
    --------
    >>> from raisin.encapsulation.packaging import Func
    >>> from raisin.serialization.atoms.raisin_class import serialize_func
    >>> Func(lambda x: x**2)
    Func(<lambda>)
    >>> data = b''.join(serialize_func(_, compact=True))
    >>> type(data)
    <class 'bytes'>
    >>> data != b''
    True
    >>>
    """
    yield HEADER['func'][compact]
    yield from obj.__getstate__(as_iter=True, compresslevel=compact, **kwds)


def deserialize_func(pack, gen, **kwds):
    r"""
    ** Deserialize ``raisin.encapsulation.packaging.Func``. **

    Examples
    --------
    >>> from raisin.encapsulation.packaging import Func
    >>> from raisin.serialization.atoms.raisin_class import serialize_func, deserialize_func
    >>> Func(lambda x: x**2)
    Func(<lambda>)
    >>> data = b''.join(serialize_func(_, compact=True))
    >>> deserialize_func(pack=data, gen=[])
    Func(<lambda>)
    >>>
    """
    from raisin.encapsulation.packaging import Func
    return _init_class(Func, to_gen(pack=pack, gen=gen), **kwds)


def serialize_task(obj, compact, **kwds):
    r"""
    ** Serialize the ``raisin.encapsulation.packaging.Task``. **

    Examples
    --------
    >>> from raisin.encapsulation.packaging import Argument, Func, Task
    >>> from raisin.serialization.atoms.raisin_class import serialize_task
    >>> arg = Argument(0)
    >>> func = Func(lambda x : x**2)
    >>> task = Task(func.__hash__(), (arg.__hash__(),))
    >>> b''.join(serialize_task(task, compact=True))
    b"'\x0e\x8a\x01\x03~\xc3\xe6\x03#\xe6\xab\x90\x82\x00\x0e\x8c\x01\x8a\x01\x03w\x06z\xe3!\xd2A\xdc"
    >>>
    """
    yield HEADER['task'][compact]
    yield from obj.__getstate__(as_iter=True, compresslevel=compact, **kwds)


def deserialize_task(pack, gen, **kwds):
    r"""
    ** Deserialize ``raisin.encapsulation.packaging.Task``. **

    Examples
    --------
    >>> from raisin.serialization.atoms.raisin_class import deserialize_task
    >>> pack = b'\x0e\x8a\x01\x03~\xc3\xe6\x03#\xe6\xab\x90\x82\x00\x0e\x8c\x01\x8a\x01\x03w\x06z\xe3!\xd2A\xdc'
    >>> deserialize_task(pack=pack, gen=[])
    Task(9134397370399107984, (8576677656317280732,))
    >>>
    """
    from raisin.encapsulation.packaging import Task
    return _init_class(Task, to_gen(pack=pack, gen=gen), **kwds)


def serialize_result(obj, compact, **kwds):
    r"""
    ** Serialize the ``raisin.encapsulation.packaging.Result``. **

    Examples
    --------
    >>> from raisin.encapsulation.packaging import Result
    >>> from raisin.serialization.atoms.raisin_class import serialize_result
    >>> res = Result()
    >>> b''.join(serialize_result(res, compact=True))
    b'('
    >>>
    """
    yield HEADER['result'][compact]
    yield from obj.__getstate__(as_iter=True, compresslevel=compact, **kwds)


def deserialize_result(pack, gen, **kwds):
    r"""
    ** Deserialize ``raisin.encapsulation.packaging.Result``. **

    Examples
    --------
    >>> from raisin.serialization.atoms.raisin_class import deserialize_result
    >>> deserialize_result(pack=b'', gen=[])
    Result()
    >>>
    """
    from raisin.encapsulation.packaging import Result
    return _init_class(Result, to_gen(pack=pack, gen=gen), **kwds)



#!/usr/bin/env python3

"""
** Configuration file for pypi. **
----------------------------------

#!/bin/bash

python3 setup.py check -s -r
python3 -m pip install --user --upgrade setuptools wheel
python3 setup.py sdist bdist_wheel
python3 -m pip install --user --upgrade twine
python3 -m twine upload --repository pypi dist/*
twine check dist/*
"""

import setuptools

with open('README.rst', 'r', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='raisin',
    version='0.1.3',
    author='Robin RICHARD (robinechuca)',
    author_email='raisin@ecomail.fr',
    description='Simple parallel, distributed and cluster computing',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://framagit.org/robinechuca/raisin/-/blob/master/README.rst',
    packages=setuptools.find_packages(),
    install_requires=['dill'],
    # install_requires=['pycryptodomex', 'cloudpickle',
    #     'unidecode', 'dirhash', 'psutil', 'requests', 'numpy',
    #     'getmac'], # Ces paquets serons installes d'office.
    # extras_require={
    #     'calculation': ['sympy', 'giacpy', 'numpy'],
    #     'tools': ['psutil>=5.1', 'regex', 'cloudpickle',
    #         'unidecode', 'dirhash', 'getmac'],
    #     'graphical': ['tkinter', 'matplotlib', 'graphviz'],
    #     'security': ['pycryptodomex', 'requests']},
    extras_require={},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Natural Language :: English',
        'Natural Language :: French',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: BSD',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Clustering',
        'Topic :: Internet',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Topic :: System :: Power (UPS)'],
    keywords=['parallel', 'distributed', 'cluster computing', 'serialization', 'serialize'],
    python_requires='>=3.6',
    project_urls={
        'Source Repository': 'https://framagit.org/robinechuca/raisin/-/tree/master/raisin',
        # 'Bug Tracker': 'https://github.com/engineerjoe440/ElectricPy/issues',
        'Documentation': 'http://raisin-docs.ddns.net',
        # 'Packaging tutorial': 'https://packaging.python.org/tutorials/distributing-packages/',
        },
    include_package_data=True,
)



#!/usr/bin/env python3

"""
** Server TCP. **
-----------------

It is a tcp server listening to several clients. Unlike the client,
there must be only one instance per machine. The server must be executed
in a totally different process than the client.
It is the pillar that allows the processes to communicate with each other.
That it is via the internal sockets within the same machine
or via sockets tcp through the Internet network.
"""


import queue
import socket
import threading

from raisin.communication.abstraction import SocketAbstractConn
from raisin.communication.handler import Handler


__pdoc__ = {
    'Server.__del__': True,
    'Server.__enter__': True,
    'Server.__exit__': True,
    'Server.__repr__': True,
    'Server.__str__': True}


class BaseServer(threading.Thread):
    """
    ** TCP server. **

    This server is both able to listen in ipv4 and ipv6.

    Attributes
    ----------
    port : int
        The listening port number.
    clients : list
        It is in this attribute that all clients that are
        communicating with the server are stored.
    tcp_socket : socket.socket
        The socket object on which the server will listen for incoming requests.
    """

    def __init__(self, port):
        """
        Parameters
        ----------
        port : int
            The port to listen on.
        """
        threading.Thread.__init__(self)
        self.port = port
        self.clients = []
        self.tcp_socket = BaseServer._init_tcp_socket(self.port)
        self.tcp_socket.listen()

        self._shutdown = False  # The flag becomes True if you have to die.
        self._request_queue = queue.Queue()  # Contains new clients.

    @staticmethod
    def _init_tcp_socket(port):
        """
        ** Help for the ``BaseServer.__init__``. **

        Paremeters
        ----------
        port : int
            The port to listen on.

        Returns
        -------
        tcp_socket : socket.socket
            The TCP socket ready to listen.

        Examples
        --------
        >>> from raisin.communication.server import BaseServer
        >>> tcp_socket = BaseServer._init_tcp_socket(9999)
        >>> type(tcp_socket)
        <class 'socket.socket'>
        >>> tcp_socket.close()
        >>>
        """
        if hasattr(socket, 'has_dualstack_ipv6'):  # only supported for python >= 3.8
            if socket.has_dualstack_ipv6():
                return socket.create_server(
                    ('', port), family=socket.AF_INET6, dualstack_ipv6=True, reuse_port=True
                )
            return socket.create_server(('', port), reuse_port=True)
        tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        tcp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        tcp_socket.bind(('', port))
        return tcp_socket

    def _get(self, request):
        """
        ** Prepare the client to pass the verification stape. **

        Parameters
        ----------
        request : tuple
            The tuple (conn, address) returned by the *socket.accept()* method.
        """
        conn, _ = request
        self._request_queue.put(
            {'stape': 'verify', 'abstract_conn': SocketAbstractConn(conn=conn)}
        )

    def _verify(self, abstract_conn):
        """
        ** Ensures the caller is welcome. **

        Parameters
        ----------
        abstract_conn : AbstractConn
            The connection to be checked.

        Notes
        -----
        For the moment, accept directly without making any verification.
        """
        self._request_queue.put({'stape': 'process', 'abstract_conn': abstract_conn})

    def _process(self, abstract_conn):
        """
        ** Definitely adds the client to the server. **

        Parameters
        ----------
        abstract_conn : AbstractConn
            The connection to be checked.
        """
        client_handler = Handler(abstract_conn)
        client_handler.daemon = True
        client_handler.start()
        self.clients.append(client_handler)

    def handle_request(self, request):
        """
        ** Process a single request. **

        This method creates a thread or a process in order to return immediately.
        In fact this method is called by the ``BaseServer.serve_forever`` method.

        Parameters
        ----------
        request : dict
            A client request that can be at different stages:

            - *get* : This is the client's first attempt to connect.
            - *verify* : Verification the client is welcome.
            - *process* : Creates an environment that allows dialogue with the client.
        """
        if request['stape'] == 'get':
            return self._get(request['request'])
        if request['stape'] == 'verify':
            return self._verify(request['abstract_conn'])
        if request['stape'] == 'process':
            return self._process(request['abstract_conn'])
        raise NameError(f"this step has nothing to do here: {request['stape']}")

    def run(self):
        """
        ** Start the server asynchronously. **

        Should not be called as is. It is the call of the
        *start* method that executes run.
        It is an alias to ``BaseServer.serve_forever``.
        """
        self.serve_forever()

    def serve_forever(self, poll_interval=0.5):
        """
        ** Handle requests until an explicit ``BaseServer.shutdown`` request. **

        Poll for shutdown every poll_interval seconds.
        """
        def listen(self):
            self.tcp_socket.settimeout(poll_interval) # avoids creation of zombies
            while not self._shutdown:
                try:
                    self._request_queue.put({'stape': 'get', 'request': self.tcp_socket.accept()})
                except socket.timeout:
                    continue
                except OSError:
                    break

        thread = threading.Thread(target=listen, args=(self,))
        thread.daemon = True
        thread.start()
        while not self._shutdown:
            try:
                self.handle_request(self._request_queue.get(timeout=poll_interval))
            except queue.Empty:
                continue
        thread.join()

    def shutdown(self):
        """
        Tell the ``BaseServer.serve_forever`` loop to stop and wait until it does.
        ``BaseServer.shutdown`` must be called while ``BaseServer.serve_forever``
        is running in a different thread otherwise it will deadlock.
        """
        self.server_close()
        while self.is_alive():
            continue
        for client in self.clients:
            client.join()

    def server_close(self):
        """
        ** Clean up the server. **

        Should not be called if the server is encapsulated
        in a context manager (*with* statement).
        Can be called several times.
        """
        self._shutdown = True
        try:
            self.tcp_socket.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        self.tcp_socket.close()
        for client in self.clients:
            client.handler_close()


class Server(BaseServer):
    """
    ** Enables you to enrich the ``raisin.communication.server.BaseServer``. **

    Examples
    --------
    >>> import time
    >>> from raisin.communication.server import Server
    >>> with Server(9999) as server:
    ...     server.start() # the server is listening
    ...     time.sleep(1) # wait until the server is ready
    ...     server.shutdown() # stops the server
    ...
    >>>
    """

    def __del__(self):
        """
        ** Help for the garbage-collector. **
        """
        self.server_close()

    def __enter__(self):
        """
        ** Prepared for easy closing. **

        Allows you to use the *with* statement which allows
        you to set up a context manager.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        ** Stop the server. **

        Goes together with ``Server.__enter__``.
        """
        self.shutdown()

    def __repr__(self):
        """
        ** Gives a simple representation of the server. **
        """
        return f'Server({self.port})'

    def __str__(self):
        r"""
        ** Provides a complete representation of the server. **
        """
        return (
            f'TCP Server:\n'
            f'    port={self.port}\n'
            f'    tcp_socket={self.tcp_socket}\n'
            f'    clients={self.clients}')


if __name__ == '__main__':
    # Launches a server without returning the hand.
    with Server(20001) as server:
        server.serve_forever()



#!/usr/bin/env python3

"""
** Allows to serialize and deserialize small objects native to python. **
-------------------------------------------------------------------------

Each function is specialized in only 1 type of objects.
The functions are called automatically according to the name of the classes of the objects.
That's why functions must end with the exact name of the class they deal with.
"""

import struct
import sys

from raisin.serialization.constants import HEADER


def serialize_bytes(obj, compact, **_):
    r"""
    ** Serialize the bytes. **

    Examples
    --------
    >>> from raisin.serialization.atoms.small import serialize_bytes
    >>> b''.join(serialize_bytes(b'', compact=False))
    b'/bytes/'
    >>> b''.join(serialize_bytes(b'', compact=True))
    b'\t'
    >>> b''.join(serialize_bytes(bytes(10), compact=True))
    b'\t\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    >>>
    """
    yield HEADER['bytes'][compact]
    yield obj

def serialize_float(obj, compact, **_):
    r"""
    ** Serialize the floats. **

    Examples
    --------
    >>> from raisin.serialization.atoms.small import serialize_float
    >>> b''.join(serialize_float(.0, compact=False))
    b'/float/0.0'
    >>> b''.join(serialize_float(.0, compact=True))
    b'\x06\xa8'
    >>> b''.join(serialize_float(2e16, compact=False))
    b'/float/2e+16'
    >>> b''.join(serialize_float(1.23456, compact=True))
    b'\x06\xb7\xaa.L'
    >>> b''.join(serialize_float(1.23456789123456, compact=True))
    b'\x07\xb7\xb4\xd8B\xca\xc0\xf3?'
    >>>
    """
    if compact:
        as_str = str(obj)
        if as_str.startswith('0.'):  # 0.123 -> .123
            as_str = as_str[1:]
        elif as_str.startswith('-0.'):  # -0.123 -> -.123
            as_str = as_str[0] + as_str[2:]
        elif as_str.endswith('.0'):  # 12.0 -> 12
            as_str = as_str[:-2]
            end_0 = 0  # nbr of 0 at end
            while as_str[-end_0 - 1 :] == '0' * (end_0 + 1):
                end_0 += 1
            if end_0 >= 3:
                as_str = as_str[:-end_0] + 'e' + str(end_0)
        as_str = as_str.replace('e+', 'e')

        if len(as_str) < 16:
            if len(as_str) % 2:
                as_str = ' ' + as_str
            pack = b''
            cor = {
                '0': 0,
                '1': 1,
                '2': 2,
                '3': 3,
                '4': 4,
                '5': 5,
                '6': 6,
                '7': 7,
                '8': 8,
                '9': 9,
                '-': 10,
                'e': 11,
                '.': 12,
                ' ': 13,
            }
            for i in range(0, len(as_str), 2):
                pack += bytes([cor[as_str[i + 1]] + cor[as_str[i]] * 14])
            yield HEADER['round_float'][1] + pack
        else:
            yield HEADER['normal_float'][1] + struct.pack('d', obj)
    else:
        yield HEADER['float'][0] + str(obj).encode()


def deserialize_float(pack, gen, **_):
    """
    ** Deserialize non compact floats. **

    Examples
    --------
    >>> from raisin.serialization.atoms.small import deserialize_float
    >>> deserialize_float(pack=b'0.0', gen=[])
    0.0
    >>>
    """
    return float(pack + b''.join(gen))


def deserialize_round_float(pack, gen, **_):
    r"""
    ** Deserialize non compact floats. **

    Examples
    --------
    >>> from raisin.serialization.atoms.small import deserialize_round_float
    >>> deserialize_round_float(pack=b'\xa8', gen=[])
    0.0
    >>> deserialize_round_float(pack=b'\xb7\xaa.L', gen=[])
    1.23456
    >>>
    """
    cor = {
        0: '0',
        1: '1',
        2: '2',
        3: '3',
        4: '4',
        5: '5',
        6: '6',
        7: '7',
        8: '8',
        9: '9',
        10: '-',
        11: 'e',
        12: '.',
        13: ' ',
    }
    return float(''.join(cor[octet // 14] + cor[octet % 14] for octet in (pack + b''.join(gen))))


def deserialize_normal_float(pack, gen, **_):
    r"""
    ** Deserialize non compact floats. **

    Examples
    --------
    >>> from raisin.serialization.atoms.small import deserialize_normal_float
    >>> deserialize_normal_float(pack=b'\xb7\xb4\xd8B\xca\xc0\xf3?', gen=[])
    1.23456789123456
    >>>
    """
    return struct.unpack('d', pack + b''.join(gen))[0]


def serialize_int(obj, compact, **_):
    r"""
    ** Serialize the integers. **

    Examples
    --------
    >>> from raisin.serialization.atoms.small import serialize_int
    >>> b''.join(serialize_int(0, compact=False))
    b'/small int/0'
    >>> b''.join(serialize_int(0, compact=True))
    b'\x03\x00'
    >>> b''.join(serialize_int(3**250, compact=False))
    b"/large int/\x12\xe7qO\x92Z\xe0'\xde\x89)0rQ\xed(\xe3f$\x17\xc4G0 \xc4\xa6\xfbn\xe7\xdc[\xc8\xa2%ggw\r\xa6\x1c\x914\x15\xb5\x93D\xe4\x9eqi"
    >>>
    """
    if not compact and sys.getsizeof(obj) <= 64:
        yield HEADER['small_int'][0] + str(obj).encode()
    else:
        yield HEADER['large_int'][compact] + obj.to_bytes(
            length=(8 + (obj + (obj < 0)).bit_length()) // 8, byteorder='big', signed=True
        )


def deserialize_small_int(pack, gen, **_):
    """
    ** Deserialize small integers. **

    Examples
    --------
    >>> from raisin.serialization.atoms.small import deserialize_small_int
    >>> deserialize_small_int(pack=b'0', gen=[])
    0
    >>>
    """
    return int(pack + b''.join(gen))


def deserialize_large_int(pack, gen, **_):
    r"""
    ** Deserialize large integers. **

    Examples
    --------
    >>> from raisin.serialization.atoms.small import deserialize_large_int
    >>> deserialize_large_int(pack=b'#T\xd0i\\\x88\x00\xeeY\xe26\xb2{\xafm\x7f\xab\x99\x1e\x82]\xf0'
    ...     b'n.\x80@\xe344\x86\xae\xfe\xb4!{\x90\xcd9', gen=[])
    4498196224760364601242719132174628305800834098010033971355568455673974002968757862019419449
    >>>
    """
    return int.from_bytes(pack + b''.join(gen), byteorder='big', signed=True)


def serialize_nonetype(compact, **_):
    r"""
    ** Serialize the NoneType. **

    Examples
    --------
    >>> from raisin.serialization.atoms.small import serialize_nonetype
    >>> b''.join(serialize_nonetype(False))
    b'/none/'
    >>> b''.join(serialize_nonetype(True))
    b'\x13'
    >>>
    """
    yield HEADER['null'][compact]


def deserialize_null(**_):
    """
    ** Deserialize None. **

    Examples
    --------
    >>> from raisin.serialization.atoms.small import deserialize_null
    >>> deserialize_null()
    >>>
    """
    return None



#!/usr/bin/env python3

"""
** All major constants and invariants for serialization. **
-----------------------------------------------------------
"""

BUFFER_SIZE = 8 * 1024 * 1024  # packet size in bytes
ALPHABET = [
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
    't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
    'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '0', '1', '2', '3', '4',
    '5', '6', '7', '8', '9', '!', '#', '$', '_', '&', '(', ')', '*', '+', ']', '-', '.', '/', ':',
    ';', '<', '=', '>', '?', '@', '[']  # alphabet to convert bytes in ascii str
N_SYMB = 64  # nbr of ascii char to encode *N_BYTES* : 51*log(256)/log(len(ALPHABET)) = 63.9996
N_BYTES = 51  # number of bytes to group before converting them into ascii str
HEADER = {
    'pickle': [b'/pickle/', b'\x00'],
    # 'json': [b'/json/', b'\x01'],
    'dumps': [b'0', b'\x30'],
    'small_int': [b'/small int/', b'\x02'],
    'large_int': [b'/large int/', b'\x03'],
    # 'filename': [b'/filename + file_content/', b'\x04'],
    # 'str': [b'/str/', b'\x05'],
    'round_float': [b'/round float/', b'\x06'],
    'normal_float': [b'/normal float/', b'\x07'],
    'float': [b'/float/', b'\x08'],
    'bytes': [b'/bytes/', b'\x09'],
    # 'TextIOWrapper': [b'/io.TextIOWrapper/', b'\x0a'],
    # 'BufferedReader': [b'/io.BufferedReader/', b'\x0b'],
    # 'BufferedWriter': [b'/io.BufferedWriter/', b'\x0c'],
    # 'list': [b'/list/', b'\x0d'],
    'tuple': [b'/tuple/', b'\x0e'],
    # 'dict': [b'/dict/', b'\x0f'],
    # 'complex': [b'/complex/', b'\x10'],
    # 'class': [b'/class/', b'\x11'],
    # 'callable': [b'/callable/', b'\x12'],
    'null': [b'/none/', b'\x13'],
    'true': [b'/true/', b'\x14'],
    'false': [b'/false/', b'\x15'],
    # 'set': [b'/set/', b'\x16'],
    # 'frozenset': [b'/frozenset/', b'\x17'],
    'generator': [b'/generator/', b'\x18'],
    'function': [b'/function/', b'\x19'],
    # 'rsa_bloc': [b'/rsa bloc/', b'\x20'],
    # 'aes_gen': [b'/aes gen/', b'\x21'],
    # 'rsa_gen': [b'/rsa gen/', b'\x22'],
    # 'aes_sel': [b'/aes sel/', b'\x23'],
    # 'authenticity': [b'/authenticity/', b'\x24'],
    'argument': [b'/argument/', b'\x25'],
    'func': [b'/func/', b'\x26'],
    'task': [b'/task/', b'\x27'],
    'result': [b'/result/', b'\x28']
}  # serialized data headers, (forbiden b'\x2f' = b'/')

ALPHABET2INDEX = {s: i for i, s in enumerate(ALPHABET)}  # to each ascii char, associates its index
BYTES2HEADER = {
    bytes_header: str_header
    for str_header, bytes_headers in HEADER.items()
    for bytes_header in bytes_headers
}  # to each bytes header, associates its name str
HEADERLEN = {
    size: any(h for h in BYTES2HEADER if len(h) > size)
    for size in range(max(len(h) for h in BYTES2HEADER) + 1)
}  # at each size, gives True if exists a bytes header of len > size



#!/usr/bin/env python3

"""
** Client TCP. **
-----------------

As soon as you want to connect to a server, you have
to create a client, even if it is for a small communication.
These clients are meant to be created and deleted in large numbers if needed.
"""

import socket
import threading

from raisin.communication.abstraction import SocketAbstractConn
from raisin.communication.handler import Handler


class BaseClient(threading.Thread, SocketAbstractConn):
    """
    ** TCP Client. **

    This client is both able to listen in ipv4 and ipv6.

    Attributes
    ----------
    host : str
        The ip address of the connection in ipv4 or ipv6.
        It can also be a hostname or a domain name.
    port : int
        The communication port.
    tcp_socket : socket.socket
        The tcp socket which allows low level communication.
    """

    def __init__(self, host, port):
        """
        Parameters
        ----------
        ip : str
            The server address.
        port : int
            The server's listening port.

        Raises
        ------
        ConnectionError
            If we can't connect.
        """
        threading.Thread.__init__(self)
        self.host = host
        self.port = port
        self.tcp_socket = BaseClient._init_tcp_socket(self.host, self.port)
        SocketAbstractConn.__init__(self, self.tcp_socket)

    @staticmethod
    def _init_tcp_socket(host, port):
        """
        ** Help for the ``BaseServer.__init__``. **

        Paremeters
        ----------
        port : int
            The port to listen on.

        Returns
        -------
        tcp_socket : socket.socket
            The TCP socket ready to communicate.

        Examples
        --------
        >>> from raisin.communication.server import BaseServer
        >>> tcp_socket = BaseServer._init_tcp_socket(9999)
        >>> type(tcp_socket)
        <class 'socket.socket'>
        >>> tcp_socket.close()
        >>>
        """
        tcp_socket = None
        for res in socket.getaddrinfo(host, port, socket.AF_UNSPEC, socket.SOCK_STREAM):
            family, socktype, proto, _, sockaddr = res
            try:
                tcp_socket = socket.socket(family, socktype, proto)
            except OSError:
                tcp_socket = None
                continue
            try:
                tcp_socket.connect(sockaddr)
            except OSError:
                tcp_socket.close()
                tcp_socket = None
                continue
            break
        if tcp_socket is None:
            raise ConnectionError('could not open socket')
        return tcp_socket

    def run(self):
        """
        ** Puts the client on asynchronous listening mode. **

        Should not be called as is. It is the call of the
        *start* method that executes run.
        """
        Handler(self).run()

    def shutdown(self):
        """
        Tell the ``BaseClient.run`` loop to stop and wait until it does.
        ``BaseClient.shutdown`` must be called while ``BaseClient.run``
        is running in a different thread otherwise it will deadlock.
        """
        self.client_close()
        while self.is_alive():
            continue

    def client_close(self):
        """
        ** Clean up the client. **

        Should not be called if the client is encapsulated
        in a context manager (*with* statement).
        Can be called several times.
        """
        try:
            self.tcp_socket.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        self.tcp_socket.close()


class Client(BaseClient):
    """
    ** Enables you to enrich the ``raisin.communication.client.BaseClient``. **

    Examples
    --------
    >>> import time
    >>> from raisin.communication.server import Server
    >>> from raisin.communication.client import Client
    >>> with Server(9999) as server:
    ...     server.start()
    ...     with Client(None, 9999) as client:
    ...         client.start()
    ...         time.sleep(1)
    >>>
    """

    def __del__(self):
        """
        ** Help for the garbage-collector. **
        """
        try:
            self.client_close()
        except AttributeError:
            pass

    def __enter__(self):
        """
        ** Prepared for easy closing. **

        Allows you to use the *with* statement which allows
        you to set up a context manager.
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        ** Stop the client. **

        Goes together with ``Server.__enter__``.
        """
        self.shutdown()

    def __repr__(self):
        """
        ** Gives a simple representation of the client. **
        """
        return f'Client({self.host}, {self.port})'

    def __str__(self):
        r"""
        ** Provides a complete representation of the client. **
        """
        return (
            f'TCP Client:\n'
            f'    host={self.host}\n'
            f'    port={self.port}\n'
            f'    tcp_socket={self.tcp_socket}')
